import tensorflow as tf


class UtilityVariable:
    """

    """

    @staticmethod
    def initialize_weight(shape):
        kernel = tf.truncated_normal(shape, stddev=0.1)
        return tf.Variable(kernel)

    @staticmethod
    def initialize_bias(shape):
        initial = tf.constant(0.1, shape=shape)
        return tf.Variable(initial)





import tensorflow as tf

from easy_net_tf.utility.variable import UtilityVariable


class MobileNetV2:
    def __init__(self,
                 image_in,
                 channels_in,
                 dilatation,
                 filter_size,
                 channels_out,
                 dilate_active_func,
                 depth_active_func,
                 stride=1,
                 add_residuals=False,
                 add_bias=True,
                 padding='SAME'):
        """
        :param image_in:
        :param channels_in:
        :param dilatation: if 'None', become MobileNet V1 without output activation
        :param filter_size:
        :param channels_out:
        :param dilate_active_func: active function for dilatation.
        :param depth_active_func: active function for depthwise.
        :param stride:
        :param add_residuals:
        :param add_bias: True: add; False: not add
        :param padding:
        """

        """
        prepare
        """
        self.channels_in = channels_in
        self.dilatation = dilatation
        self.filter_size = filter_size
        self.channels_out = channels_out
        self.dilate_active_func = dilate_active_func
        self.depth_active_func = depth_active_func
        self.stride = stride
        self.add_residuals = add_residuals
        self.add_bias = add_bias
        self.padding = padding

        """
        initialize filter
        """
        self.dilatation_filter, \
        self.depthwise_filter, \
        self.compress_filter, \
        self.residuals_filter, \
        self.bias = self._initialize_variable()

        """
        calculation
        """
        self.feature_map = self._calculate(image_in)

    def _initialize_variable(self):
        """
        initialize variable
        :return:
        """

        channel_multiplier = 1

        """
        dilatation_filter
        """
        if self.dilatation is None:
            dilatation_filter = None
            self.dilatation = 1
        else:
            dilatation_filter = UtilityVariable.initialize_weight(
                [1,
                 1,
                 self.channels_in,
                 self.channels_in * self.dilatation]
            )

        """
        depthwise_filter
        """
        depthwise_filter = UtilityVariable.initialize_weight(
            [self.filter_size,
             self.filter_size,
             self.channels_in * self.dilatation,
             channel_multiplier]
        )

        """
        compress_filter
        """
        compress_filter = UtilityVariable.initialize_weight(
            [1,
             1,
             self.channels_in * self.dilatation * channel_multiplier,
             self.channels_out]
        )

        """
        residuals_filter
        """
        if self.add_residuals and self.channels_in != self.channels_out:
            residuals_filter = UtilityVariable.initialize_weight(
                [1,
                 1,
                 self.channels_in,
                 self.channels_out]
            )
        else:
            residuals_filter = None

        """
        bias
        """
        if self.add_bias:
            bias = UtilityVariable.initialize_bias([self.channels_out])
        else:
            bias = None

        return dilatation_filter, \
               depthwise_filter, \
               compress_filter, \
               residuals_filter, \
               bias

    def _calculate(self,
                   image_in):
        """
        No activation function applied on output.
        :param image_in:
        :return:
        """

        """
        dilatation
        """
        if self.dilatation_filter is not None:

            dilatation_o = tf.nn.conv2d(input=image_in,
                                        filter=self.dilatation_filter,
                                        strides=[1, 1, 1, 1],
                                        padding='SAME')

            if self.depth_active_func is not None:
                dilatation_o = self.dilate_active_func(dilatation_o)

        else:
            dilatation_o = image_in

        """
        depthwise
        """
        convolution_o = tf.nn.depthwise_conv2d(input=dilatation_o,
                                               filter=self.depthwise_filter,
                                               strides=[1, self.stride, self.stride, 1],
                                               padding=self.padding)
        if self.depth_active_func is not None:
            convolution_o = self.depth_active_func(convolution_o)

        """
        compress
        """
        compress_o = tf.nn.conv2d(input=convolution_o,
                                  filter=self.compress_filter,
                                  strides=[1, 1, 1, 1],
                                  padding='SAME')

        """
        residuals block
        """
        if self.add_residuals is True:
            if self.residuals_filter is not None:
                image_in = tf.nn.conv2d(input=image_in,
                                        filter=self.residuals_filter,
                                        strides=[1, 1, 1, 1],
                                        padding='SAME')

            result = tf.add(compress_o, image_in)
        else:
            result = compress_o

        """
        bias
        """
        if self.bias is not None:
            result += self.bias

        return result

    def get_variable(self):
        """
        :return: filters
        """
        return self.dilatation_filter, \
               self.depthwise_filter, \
               self.compress_filter

    def get_feature_map(self):
        """
        :return: image out
        """
        return self.feature_map



import tensorflow as tf


class UtilityLoss:
    @staticmethod
    def cross_entropy_loss(prediction,
                           truth,
                           selections,
                           keep_ratio,
                           name,
                           debug=False):
        """

        :param prediction:
        :param truth: ground truth category. int32, int64 only
        :param selections: a list of selected categories.
        :param keep_ratio: keep top high 0.7 of cross entropy.
        :param name: a name for tf.summary
        :param debug: True: return loss,
                            valid_index,
                            raw_cross_entropy,
                            valid_cross_entropy,
                            top_cross_entropy;
                      False: return loss
        :return:
        """

        selection_0 = selections.pop(0)

        """
        valid index
        """
        valid_index = tf.equal(truth, selection_0)

        for valid_category in selections:
            valid_index = tf.logical_or(
                valid_index,
                tf.equal(truth, valid_category)
            )

        valid_total = tf.reduce_sum(
            tf.cast(
                valid_index,
                dtype=tf.float32)
        )

        mask_b = tf.multiply(
            tf.ones_like(truth),
            selection_0
        )
        truth = tf.cast(
            tf.where(
                valid_index,
                truth,
                mask_b
            ),
            tf.int32
        )

        """
        cross entropy
        """
        raw_cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=truth,
            logits=prediction
        )

        mask_0 = tf.zeros_like(raw_cross_entropy)

        valid_cross_entropy = tf.where(
            valid_index,
            raw_cross_entropy,
            mask_0
        )

        """
        keep ratio
        """
        keep_number = tf.cast(
            keep_ratio * valid_total,
            dtype=tf.int32
        )

        top_cross_entropy, _ = tf.nn.top_k(valid_cross_entropy, k=keep_number)

        """
        loss
        """
        loss = tf.reduce_mean(top_cross_entropy)

        tf.summary.scalar(name=name,
                          tensor=loss)

        if debug:
            return loss, valid_index, raw_cross_entropy, valid_cross_entropy, top_cross_entropy
        else:
            return loss

    @staticmethod
    def l2_loss(prediction,
                truth,
                categories,
                selections,
                keep_ratio,
                name,
                debug=False):
        """

        :param prediction:
        :param truth:
        :param categories: category label for each sample
        :param selections: a list of selected categories.
        :param keep_ratio: keep top high 0.7 of square error.
        :param name: a name for tf.summary
        :param debug: True: return loss,
                            valid_index,
                            raw_square_error,
                            valid_square_error,
                            top_square_error;
                      False: return loss
        :return:
        """

        selection_0 = selections.pop(0)

        """
        valid index
        """
        valid_index = tf.equal(categories, selection_0)

        for selection in selections:
            valid_index = tf.logical_or(
                valid_index,
                tf.equal(categories, selection)
            )

        valid_total = tf.reduce_sum(
            tf.cast(
                valid_index,
                dtype=tf.float32)
        )

        """
        square error
        """
        raw_square_error = tf.reduce_sum(
            tf.square(prediction - truth),
            axis=1
        )

        mask_0 = tf.zeros_like(raw_square_error)
        valid_square_error = tf.where(
            valid_index,
            raw_square_error,
            mask_0
        )

        """
        keep ratio
        """
        keep_num = tf.cast(
            keep_ratio * valid_total,
            dtype=tf.int32
        )

        top_square_error, _ = tf.nn.top_k(valid_square_error, k=keep_num)

        """
        loss
        """
        loss = tf.reduce_mean(top_square_error)

        tf.summary.scalar(name=name,
                          tensor=loss)

        if debug:
            return loss, valid_index, raw_square_error, valid_square_error, top_square_error
        else:
            return loss


if __name__ == '__main__':
    from easydict import EasyDict
    import numpy

    predicted = EasyDict()
    predicted.category = numpy.array([[0.9, 0.1],
                                      [0.8, 0.2],
                                      [0.7, 0.3],
                                      [0.6, 0.4],
                                      [0.5, 0.5],
                                      [0.4, 0.6],
                                      [0.3, 0.7],
                                      [0.2, 0.8],
                                      [0.1, 0.9]])
    predicted.offset = numpy.array([[0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1],
                                    [0.1, 0.1, 0.1, 0.1]])

    ground_truth = EasyDict()
    ground_truth.category = numpy.array([1,
                                         2,
                                         3,
                                         1,
                                         2,
                                         1,
                                         0,
                                         2,
                                         0])
    ground_truth.offset = numpy.array([[0.0, 0.1, 0.1, 0.1],
                                       [0.0, 0.0, 0.1, 0.1],
                                       [0.0, 0.0, 0.0, 0.1],
                                       [0.0, 0.0, 0.0, 0.0],
                                       [0.3, 0.1, 0.1, 0.1],
                                       [0.3, 0.3, 0.1, 0.1],
                                       [0.3, 0.3, 0.3, 0.1],
                                       [0.3, 0.3, 0.3, 0.3],
                                       [0.1, 0.5, 0.1, 0.1]])

    cross_entropy_op = UtilityLoss.cross_entropy_loss(prediction=predicted.category,
                                                      truth=ground_truth.category,
                                                      selections=[0, 1],
                                                      keep_ratio=0.8,
                                                      name='category',
                                                      debug=True)

    offset_loss_op = UtilityLoss.l2_loss(prediction=predicted.offset,
                                         truth=ground_truth.offset,
                                         categories=ground_truth.category,
                                         selections=[2],
                                         keep_ratio=1,
                                         name='offset',
                                         debug=True)

    with tf.Session() as sess:
        cross_entropy_ = sess.run(cross_entropy_op)
        offset_loss_ = sess.run(offset_loss_op)

        print(cross_entropy_)
        print(offset_loss_)



import tensorflow as tf
from easy_net_tf.utility.variable import UtilityVariable


class FullyConnectNet:
    def __init__(self,
                 features_in,
                 nodes_in,
                 nodes_out):

        """

        :param features_in:
        :param nodes_in: input nodes number
        :param nodes_out: output nodes number
        """

        """
        initialize variable
        """
        self.weight, \
        self.bias = self._initialize_variable(nodes_in=nodes_in,
                                              nodes_out=nodes_out)

        """
        calculate
        """
        self.features_out = self._calculate(features_in=features_in)

    @staticmethod
    def _initialize_variable(nodes_in, nodes_out):
        """

        :param nodes_in: input nodes number
        :param nodes_out: output nodes number
        :return:
        """
        weight = UtilityVariable.initialize_weight([nodes_in, nodes_out])
        bias = UtilityVariable.initialize_bias([nodes_out])
        return weight, bias

    def _calculate(self, features_in):
        """

        :param features_in:
        :return: features
        """
        if self.bias is None:
            features_out = tf.matmul(features_in, self.weight)
        else:
            features_out = tf.matmul(features_in, self.weight) + self.bias
        return features_out

    def get_variable(self):
        """

        :return: weight, bias
        """
        return self.weight, self.bias

    def get_features(self):
        """

        :return: features
        """
        return self.features_out



import numpy


class UtilityLandmark:

    @staticmethod
    def normalize(batch_rectangle, landmark):
        """

        :param batch_rectangle: [batch, 4]
        :param landmark: [2n]
        :return: narry [batch, 2n]
        """

        """
        prepare
        """
        batch, _ = numpy.shape(batch_rectangle)

        landmark = numpy.reshape(landmark, [-1, 2])
        n_landmark = list()

        """
        normalize landmark
        """
        for rectangle in batch_rectangle:
            for x, y in landmark:
                n_landmark.append(
                    (x - rectangle[0]) / (rectangle[2] - rectangle[0] + 1)
                )

                n_landmark.append(
                    (y - rectangle[1]) / (rectangle[3] - rectangle[1] + 1)
                )

        """
        reshape
        """
        n_landmark = numpy.reshape(
            numpy.array(
                n_landmark,
                dtype=numpy.float32
            ),
            [batch, -1]
        )

        return n_landmark

    @staticmethod
    def regress(batch_box, batch_landmark):
        """
        calibrate landmarks to real scale
        :param batch_box: [batch, 4]
        :param batch_landmark: [batch, 10]
        :return:
        """
        boxes_height = batch_box[:, 2] - batch_box[:, 0] + 1
        boxes_width = batch_box[:, 3] - batch_box[:, 1] + 1

        batch_landmark[:, 0] = batch_landmark[:, 0] * boxes_width + batch_box[:, 0]
        batch_landmark[:, 1] = batch_landmark[:, 1] * boxes_height + batch_box[:, 1]

        batch_landmark[:, 2] = batch_landmark[:, 2] * boxes_width + batch_box[:, 0]
        batch_landmark[:, 3] = batch_landmark[:, 3] * boxes_height + batch_box[:, 1]

        batch_landmark[:, 4] = batch_landmark[:, 4] * boxes_width + batch_box[:, 0]
        batch_landmark[:, 5] = batch_landmark[:, 5] * boxes_height + batch_box[:, 1]

        batch_landmark[:, 6] = batch_landmark[:, 6] * boxes_width + batch_box[:, 0]
        batch_landmark[:, 7] = batch_landmark[:, 7] * boxes_height + batch_box[:, 1]

        batch_landmark[:, 8] = batch_landmark[:, 8] * boxes_width + batch_box[:, 0]
        batch_landmark[:, 9] = batch_landmark[:, 9] * boxes_height + batch_box[:, 1]

        return batch_landmark

    @staticmethod
    def align_face(image, landmarks, template):
        image_height, image_width, _ = image.shape

        landmarks = numpy.reshape(landmarks, [-1, 2])
        template = numpy.reshape(template, [-1, 2])

        # maybe it is unnecessary, wait for proving!!!
        for index, _ in enumerate(landmarks):
            landmarks[index][0] *= image_width
            landmarks[index][1] *= image_height

            template[index][0] *= image_width
            template[index][1] *= image_height

        mat = cv2.estimateRigidTransform(src=landmarks,
                                         dst=template,
                                         fullAffine=True)

        align_image = cv2.warpAffine(src=image,
                                     M=mat,
                                     dsize=(image_height, image_width))

        return align_image



import numpy

from easy_net_tf.utility.image import UtilityImage


class UtilitySlideWindow:

    @staticmethod
    def slide_window(image, window_size, stride):
        image_height, image_width, channels = image.shape

        x_1 = 0
        y_1 = 0
        x_2 = window_size - 1
        y_2 = window_size - 1

        if x_2 >= image_width or y_2 >= image_height:
            return None, None

        box_list = list()
        image_list = list()

        while y_2 < image_height - 1:

            while x_2 < image_width - 1:
                sub_image = UtilityImage.cut_out_rectangle(image_in=image,
                                                           rectangle=[x_1, y_1, x_2, y_2])

                box_list.append([x_1, y_1, x_2, y_2])
                image_list.append(sub_image)

                # go to next column
                x_1 += stride
                x_2 += stride

            # the last column
            x_1 = image_width - window_size
            x_2 = image_width - 1

            sub_image = UtilityImage.cut_out_rectangle(image_in=image,
                                                       rectangle=[x_1, y_1, x_2, y_2])
            box_list.append([x_1, y_1, x_2, y_2])
            image_list.append(sub_image)

            # back to the first column
            x_1 = 0
            x_2 = window_size - 1
            # go to next row
            y_1 += stride
            y_2 += stride

        # the last row
        y_1 = image_height - window_size
        y_2 = image_height - 1
        while x_2 < image_width - 1:
            sub_image = UtilityImage.cut_out_rectangle(image_in=image,
                                                       rectangle=[x_1, y_1, x_2, y_2])
            box_list.append([x_1, y_1, x_2, y_2])
            image_list.append(sub_image)

            # go to next column
            x_1 += stride
            x_2 += stride

        # the last column
        x_1 = image_width - window_size
        x_2 = image_width - 1

        sub_image = UtilityImage.cut_out_rectangle(image_in=image,
                                                   rectangle=[x_1, y_1, x_2, y_2])
        box_list.append([x_1, y_1, x_2, y_2])
        image_list.append(sub_image)

        images = None
        for image in image_list:

            image = numpy.reshape(image, [1, window_size, window_size, channels])
            if images is None:
                images = image
            else:
                images = numpy.vstack((images, image))

        boxes = numpy.array(box_list)

        return images, boxes



import tensorflow as tf

from easy_net_tf.utility.variable import UtilityVariable


class CNN:
    def __init__(self,
                 image_in,
                 channels_in,
                 filter_size,
                 channels_out,
                 stride,
                 add_residuals=False,
                 add_bias=True,
                 padding='SAME'):
        """
        :param image_in:
        :param channels_in:
        :param filter_size:
        :param channels_out:
        :param stride:
        :param add_residuals:
        :param add_bias: True: add; False: not add
        :param padding:
        """

        """
        prepare
        """
        self.channels_in = channels_in
        self.filter_size = filter_size
        self.channels_out = channels_out
        self.stride = stride
        self.add_residuals = add_residuals
        self.add_bias = add_bias
        self.padding = padding

        """
        initialize filter
        """
        self.convolution_filter, \
        self.residuals_filter, \
        self.bias = self._initialize_variable()

        """
        calculation
        """
        self.feature_map = self._calculate(image_in)

    def _initialize_variable(self):
        """
        initialize variable
        :return:
        """

        """
        filter
        """
        convolution_filter = UtilityVariable.initialize_weight(
            [self.filter_size,
             self.filter_size,
             self.channels_in,
             self.channels_out]
        )

        """
        residuals_filter
        """
        if self.add_residuals and self.channels_in != self.channels_out:
            residuals_filter = UtilityVariable.initialize_weight(
                [1,
                 1,
                 self.channels_in,
                 self.channels_out]
            )
        else:
            residuals_filter = None

        """
        bias
        """
        if self.add_bias:
            bias = UtilityVariable.initialize_bias([self.channels_out])
        else:
            bias = None

        return convolution_filter, residuals_filter, bias

    def _calculate(self,
                   image):
        """
        No activation function applied on output.
        :param image:
        :return:
        """

        """
        convolution
        """
        convolution_o = tf.nn.conv2d(input=image,
                                     filter=self.convolution_filter,
                                     strides=[1, self.stride, self.stride, 1],
                                     padding=self.padding)

        """
        residuals block
        """
        if self.add_residuals is True:
            if self.residuals_filter is not None:
                image = tf.nn.conv2d(input=image,
                                     filter=self.residuals_filter,
                                     strides=[1, 1, 1, 1],
                                     padding='SAME')

            map_out = tf.add(convolution_o, image)
        else:
            map_out = convolution_o

        """
        bias
        """
        if self.bias is not None:
            map_out += self.bias

        return map_out

    def get_variable(self):
        """
        :return:
        """
        return self.convolution_filter, \
               self.residuals_filter, \
               self.bias

    def get_feature_map(self):
        """
        :return:
        """
        return self.feature_map



import cv2
import numpy

from pathlib import Path


class UtilityImage:
    NORMALIZE_DEFAULT = 'default'
    NORMALIZE_EACH_CHANNEL = 'normalize each channel to [low,high]'
    NORMALIZE_ALL_CHANNEL = 'normalize all channel to [low,high]'

    @staticmethod
    def normalize(image_in, low, high, mode=NORMALIZE_DEFAULT):
        """
        
        :param image_in: 
        :param low: 
        :param high: 
        :param mode: 
        :return: 
        """
        image_out = numpy.copy(image_in)

        if mode == UtilityImage.NORMALIZE_DEFAULT:
            image_out = (image_out - 127.5) / 128

        # normalize each channel separately to [-1, 1]
        elif mode == UtilityImage.NORMALIZE_EACH_CHANNEL:
            height, width, channels = image_out.shape
            for channel in range(channels):
                maximum = numpy.max(image_out[:, :, channel])
                minimum = numpy.min(image_out[:, :, channel])
                image_out[:, :, channel] = (
                                                   (image_out[:, :, channel] - minimum) / (
                                               numpy.clip(a=maximum - minimum, a_min=1e-10, a_max=None)
                                           )
                                           ) * (high - low) + low

        # normalize all channels together to [-1, 1]
        elif mode == UtilityImage.NORMALIZE_ALL_CHANNEL:
            maximum = numpy.max(image_out)
            minimum = numpy.min(image_out)
            image_out = (
                                (image_out - minimum) / (numpy.clip(a=maximum - minimum, a_min=1e-10, a_max=None)
                                                         )
                        ) * (high - low) + low

        return image_out

    @staticmethod
    def resize(image_in, scale):
        image_out = numpy.copy(image_in)

        height, width, channels = image_out.shape
        new_height = int(height * scale)
        new_width = int(width * scale)

        image_out = cv2.resize(image_out,
                               (new_width, new_height),
                               interpolation=cv2.INTER_LINEAR)
        return image_out

    @staticmethod
    def rotate(image_in, center, angle, landmark_in=None):
        """

        :param image_in:
        :param center: rotation center
        :param landmark_in:
        :param angle:
        :return:
        """
        """
        image rotate
        """
        if image_in is None:
            return None, None

        image_out = numpy.copy(image_in)

        rotate_mat = cv2.getRotationMatrix2D(center, angle, 1)

        image_out = cv2.warpAffine(image_out, rotate_mat, (image_out.shape[1], image_out.shape[0]))

        """
        landmark rotate
        """
        if landmark_in is None:
            return image_out

        landmark_in = numpy.array(landmark_in,
                                  dtype=numpy.float32).reshape(-1, 2)
        landmark_out = list()

        for x, y in landmark_in:
            landmark_out.append(rotate_mat[0][0] * x + rotate_mat[0][1] * y + rotate_mat[0][2])
            landmark_out.append(rotate_mat[1][0] * x + rotate_mat[1][1] * y + rotate_mat[1][2])

        return image_out, landmark_out

    @staticmethod
    def horizontally_flip(image_in,
                          landmarks_in=None):
        """

        :param image_in:
        :param landmarks_in:
        :return:
        """

        """
        image flip
        """
        if image_in is None:
            print('Error: '
                  '[UtilityImage.horizontally_flip] '
                  'image_in cant be None')
            return

        image_out = numpy.copy(image_in)
        image_out = cv2.flip(image_out, 1)

        """
        landmark flip
        """
        if landmarks_in is not None:
            landmarks_out = numpy.copy(landmarks_in)
            landmarks_out = numpy.reshape(landmarks_out, [-1, 10])

            for landmark in landmarks_out:

                for i in range(10):
                    if i % 2 == 0:
                        # horizontally flip
                        landmark[i] = 1.0 - landmark[i]

                landmark[0], landmark[2] = landmark[2], landmark[0]
                landmark[1], landmark[3] = landmark[3], landmark[1]

                landmark[6], landmark[8] = landmark[8], landmark[6]
                landmark[7], landmark[9] = landmark[9], landmark[7]

            landmarks_out = numpy.reshape(landmarks_out, [-1])
        else:
            landmarks_out = None

        return image_out, landmarks_out

    @staticmethod
    def vertical_flip(image_in, landmark_in):
        if image_in is None:
            return None, None
        image_out = numpy.copy(image_in)
        image_out = cv2.flip(image_out, 0)

        for i in range(numpy.size(landmark_in)):
            if i % 2 == 1:
                # vertical flip
                landmark_in[i] = 1.0 - landmark_in[i]

        return image_out, landmark_in

    @staticmethod
    def draw_rectangle(image, boxes, color):
        """
        draw bounding boxes on image
        :param image:
        :param boxes: [x1, y1, x2, y2]
        :param color: (B, G, R)
        :return:
        """
        height, width, _ = image.shape

        boxes = numpy.reshape(boxes, [-1, 4])
        for box in boxes:
            x_1 = int(numpy.maximum(box[0], 0))
            y_1 = int(numpy.maximum(box[1], 0))
            x_2 = int(numpy.minimum(box[2], width - 1))
            y_2 = int(numpy.minimum(box[3], height - 1))

            cv2.rectangle(image,
                          (x_1, y_1),
                          (x_2, y_2),
                          color,
                          2)

        return image

    @staticmethod
    def draw_points(image, landmarks, color):
        """
        draw landmarks on image
        :param image:
        :param landmarks: shape:[None, 2]
        :param color:
        :return:
        """

        for x, y in landmarks:
            cv2.circle(img=image,
                       center=(x, y),
                       radius=2,
                       color=color,
                       thickness=-1)

        return image

    @staticmethod
    def add_coordinate(batch_image,
                       add_r,
                       normalize,
                       debug=False):
        """

        :param batch_image: a batch of images
        :param add_r: r coordinate=sqrt((x-width/2)^2+(y-height/2)^2)
        :param normalize:
        :param debug:
        :return:
        """

        """
        prepare
        """
        image_out = numpy.copy(batch_image)
        batch, height, width, channels = numpy.shape(image_out)

        """
        x coordinate
        """
        x_ones = numpy.ones(
            [batch, height],
            dtype=numpy.int32
        )

        x_ones = numpy.expand_dims(x_ones, -1)

        x_range = numpy.tile(
            numpy.expand_dims(
                numpy.arange(width),
                0
            ),
            [batch, 1]
        )

        x_range = numpy.expand_dims(x_range, 1)
        x_channel = numpy.matmul(x_ones, x_range)
        x_channel = numpy.expand_dims(x_channel, -1)

        """
        y coordinate
        """
        y_ones = numpy.ones(
            [batch, width],
            dtype=numpy.int32
        )
        y_ones = numpy.expand_dims(y_ones, 1)
        y_range = numpy.tile(
            numpy.expand_dims(
                numpy.arange(height),
                0
            ),
            [batch, 1]
        )
        y_range = numpy.expand_dims(y_range, -1)
        y_channel = numpy.matmul(y_range, y_ones)
        y_channel = numpy.expand_dims(y_channel, -1)

        """
        r coordinate
        """
        r_channel = numpy.sqrt(
            numpy.square(x_channel - (width - 1) / 2.0)
            + numpy.square(y_channel - (height - 1) / 2.0)
        )

        """
        normalize
        """
        if normalize:
            x_channel = x_channel / (width - 1.0) * 2.0 - 1.0
            y_channel = y_channel / (height - 1.0) * 2.0 - 1.0

            maximum = numpy.max(r_channel)
            minimum = numpy.min(r_channel)
            r_channel = ((r_channel - minimum) / (maximum - minimum)) * 2.0 - 1.0

        """
        concatenate
        """
        image_out = numpy.concatenate(
            (image_out, x_channel, y_channel),
            axis=-1
        )
        if add_r:
            if debug:
                print('Log: '
                      '[%s.%s] '
                      'add r coordinate.' % (UtilityImage.__name__,
                                             UtilityImage.add_coordinate.__name__))
            image_out = numpy.concatenate(
                (image_out, r_channel),
                axis=-1
            )

        return image_out

    VALID = 'VALID'
    SAME = 'SAME'

    @staticmethod
    def cut_out_rectangle(image_in,
                          rectangle,
                          size_out=None,
                          padding=VALID):
        """

        :param image_in:
        :param rectangle: [x1, y1, x2, y2]
        :param size_out: resize size
        :param padding: 'VALID': output valid image area; 'SAME': padding '0' if rectangle exceeds image
        :return:
        """

        image_copy = numpy.copy(image_in)

        if padding == UtilityImage.VALID:
            image_out = image_copy[rectangle[1]:(rectangle[3] + 1), rectangle[0]:(rectangle[2] + 1), :]

        elif padding == UtilityImage.SAME:
            height, width, channels = image_copy.shape

            x_1 = rectangle[0]
            y_1 = rectangle[1]
            x_2 = rectangle[2]
            y_2 = rectangle[3]

            m_height = y_2 - y_1 + 1
            m_width = x_2 - x_1 + 1

            if x_1 < 0:
                m_x_1 = -x_1
                x_1 = 0
            else:
                m_x_1 = 0

            if y_1 < 0:
                m_y_1 = -y_1
                y_1 = 0
            else:
                m_y_1 = 0

            if x_2 >= width:
                m_x_2 = m_width - 1 - (x_2 - (width - 1))
                x_2 = width - 1
            else:
                m_x_2 = m_width - 1

            if y_2 >= height:
                m_y_2 = m_height - 1 - (y_2 - (height - 1))
                y_2 = height - 1
            else:
                m_y_2 = m_height - 1

            # generate crop size mask
            image_out = numpy.zeros(shape=(m_height, m_width, channels))
            image_out[m_y_1:m_y_2 + 1, m_x_1:m_x_2 + 1, :] = image_copy[y_1:y_2 + 1, x_1:x_2 + 1, :]

        else:
            print('Error: '
                  '[%s.%s] '
                  'padding should be ''VALID'' or ''SAME''' % (UtilityImage.__name__,
                                                               UtilityImage.cut_out_rectangle.__name__))
            image_out = None

        if image_out is None:
            return
        elif size_out is not None \
                and size_out > 0:
            image_out = cv2.resize(image_out,
                                   (size_out, size_out),
                                   interpolation=cv2.INTER_LINEAR)
            return image_out

    @staticmethod
    def cut_out_rectangles(image_in,
                           rectangles,
                           size_out):
        """
        1.repair exceeding boxes to valid boxes, padding '0' in exceeding area
        2.resize to net input size;
        :param image_in:
        :param rectangles:
        :param size_out: resize size
        :return: a batch of resize cropped image
        """

        image_copy = numpy.copy(image_in)

        image_height, image_width, channels = image_copy.shape
        rectangles = numpy.around(rectangles).astype(numpy.int32)
        boxes_number = rectangles.shape[0]

        boxes_x_1 = rectangles[:, 0]
        boxes_y_1 = rectangles[:, 1]
        boxes_x_2 = rectangles[:, 2]
        boxes_y_2 = rectangles[:, 3]

        boxes_width = boxes_x_2 - boxes_x_1 + 1
        boxes_height = boxes_y_2 - boxes_y_1 + 1

        c_x_1 = numpy.zeros((rectangles.shape[0],), numpy.int32)
        c_y_1 = numpy.zeros((rectangles.shape[0],), numpy.int32)
        c_x_2 = boxes_width.copy() - 1
        c_y_2 = boxes_height.copy() - 1

        index = numpy.where(numpy.greater(boxes_x_2, image_width - 1))[0]
        if numpy.size(index) > 0:
            c_x_2[index] = (boxes_width[index] - 1) - (boxes_x_2[index] - (image_width - 1))
            boxes_x_2[index] = image_width - 1

        index = numpy.where(boxes_y_2 > image_height - 1)
        if numpy.size(index) > 0:
            c_y_2[index] = (boxes_height[index] - 1) - (boxes_y_2[index] - (image_height - 1))
            boxes_y_2[index] = image_height - 1

        index = numpy.where(boxes_x_1 < 0)
        if numpy.size(index) > 0:
            c_x_1[index] = 0 - boxes_x_1[index]
            boxes_x_1[index] = 0

        index = numpy.where(boxes_y_1 < 0)
        if numpy.size(index) > 0:
            c_y_1[index] = 0 - boxes_y_1[index]
            boxes_y_1[index] = 0

        images_out = numpy.zeros((boxes_number,
                                  size_out,
                                  size_out,
                                  channels),
                                 dtype=numpy.float32)

        for i in range(boxes_number):
            container = numpy.zeros((boxes_height[i],
                                     boxes_width[i],
                                     channels),
                                    dtype=numpy.float32)
            container[c_y_1[i]:c_y_2[i] + 1, c_x_1[i]:c_x_2[i] + 1, :] = image_copy[boxes_y_1[i]:boxes_y_2[i] + 1,
                                                                         boxes_x_1[i]:boxes_x_2[i] + 1, :]
            images_out[i, :, :, :] = cv2.resize(container, (size_out, size_out))

        return images_out

    @staticmethod
    def save_image(image_in,
                   category,
                   save_dir,
                   index,
                   stride,
                   landmarks=None,
                   offsets=None):
        """

        :param image_in:
        :param category: a list(negative:0, positive:1, partial:-1, landmark:-2).
        :param save_dir: directory of saving specific category
        :param index: image index
        :param stride: [images save stride, label save stride]
        :param landmarks:
        :param offsets:
        :return:
        """

        if image_in is None:
            return False

        if landmarks is not None \
                and offsets is not None:
            print('Error: landmark and offset can not be not None in the same time')
            return False

        """
        save image
        """
        image_folder = Path('%d' % int(index / stride[0]))
        image_name = '%d.jpg' % index

        image_path = save_dir / 'image' / image_folder / image_name
        image_path.parent.mkdir(parents=True, exist_ok=True)

        cv2.imwrite(filename=image_path.__str__(), img=image_in)

        """
        save labels
        """
        if landmarks is not None:
            label = ' ' + ' '.join(map(str, category)) \
                    + ' ' + ' '.join(map(str, landmarks)) + '\n'

        elif offsets is not None:
            label = ' ' + ' '.join(map(str, category)) \
                    + ' ' + ' '.join(map(str, offsets)) + '\n'

        else:
            label = ' ' + ' '.join(map(str, category)) + '\n'

        filename = 'labels-%d' % int(index / stride[1])
        file_path = save_dir / 'label' / filename
        file_path.parent.mkdir(parents=True, exist_ok=True)

        content = (image_folder / image_name).__str__() + label

        with file_path.open('a') as file:
            file.write(content)

        return True



import numpy


class UtilityAnchor:
    @staticmethod
    def generate_anchors(map_height,
                         map_width,
                         stride,
                         template,
                         anchor_number):
        """
        generate anchors centre from map
        :param map_height:
        :param map_width:
        :param stride:
        :return:
        """
        map_x = numpy.arange(start=0,
                             stop=map_width,
                             dtype=numpy.float32)
        anchors_centre_x = map_x * stride + int(stride / 2)

        map_y = numpy.arange(start=0,
                             stop=map_height,
                             dtype=numpy.float32)

        anchors_centre_y = map_y * stride + int(stride / 2)

        anchors_centre_x, anchors_centre_y = numpy.meshgrid(anchors_centre_x,
                                                            anchors_centre_y)

        anchors_centre = numpy.transpose(numpy.vstack((numpy.ravel(anchors_centre_x),
                                                       numpy.ravel(anchors_centre_y),
                                                       numpy.ravel(anchors_centre_x),
                                                       numpy.ravel(anchors_centre_y))))

        anchors_centre = numpy.repeat(anchors_centre, anchor_number, axis=0)
        anchors_centre = numpy.reshape(anchors_centre, [map_height, map_width, anchor_number, 4])

        anchors = template + anchors_centre

        return anchors

    @staticmethod
    def generate_template(min_face,
                          max_face,
                          positive_border):
        width_list = list()
        face_size = min_face
        width_list.append(face_size)
        while face_size < max_face * positive_border:
            face_size /= (positive_border * 1.025)
            width_list.append(round(face_size))

        width_list.append(max_face)

        widths = numpy.array(width_list, dtype=numpy.int32)

        anchors_x_1 = numpy.round(-(widths - 1) / 2)
        anchors_y_1 = numpy.round(-(widths - 1) / 2)
        anchors_x_2 = widths + anchors_x_1 - 1
        anchors_y_2 = widths + anchors_y_1 - 1

        template = numpy.transpose(numpy.array([anchors_x_1,
                                                anchors_y_1,
                                                anchors_x_2,
                                                anchors_y_2],
                                               dtype=numpy.int32))
        total = len(width_list)

        return template, total, widths, widths



import random
import os


class UtilityFile:
    """

    """

    @staticmethod
    def line_generator(file_path):
        with file_path.open('r') as file:
            while True:
                line = file.readline()
                if line:
                    yield line
                else:
                    return

    @staticmethod
    def line_shuffle(file_path):
        # read
        with file_path.open('r') as file:
            info_list = file.readlines()

        # shuffle
        random.shuffle(info_list)

        # write
        with file_path.open('w') as file:
            for info in info_list:
                file.write(info)

    @staticmethod
    def get_files(files_dir, shuffle):

        file_list = os.listdir(files_dir.__str__())

        if shuffle:
            random.shuffle(file_list)

        return file_list

    @staticmethod
    def get_lines(file_path, shuffle):

        print(file_path.__str__())

        if shuffle:
            UtilityFile.line_shuffle(file_path)

        return UtilityFile.line_generator(file_path)



import setuptools

with open('README.md', 'r') as file:
    long_description = file.read()

setuptools.setup(
    name="easy_net_tf",
    version="0.0.13",
    author="Ye Hang'yang",
    author_email="yehangyang@qq.com",
    description="A tensorflow-based network utility lib.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="MIT",
    url="https://gitlab.espressif.cn:6688/yhy/Network",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: Unix",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import numpy

from easy_net_tf.utility.file import UtilityFile


class UtilityBounding:
    CIRCLE = 'circle'
    RECTANGLE = 'rectangle'

    @staticmethod
    def rectangle_2_circle(r_file_path,
                           s_filename,
                           debug=False):
        """

        :param r_file_path:
        :param s_filename:
        :param debug:
        :return:
        """
        info_generator = UtilityFile.line_generator(r_file_path)
        s_file_path = r_file_path.parent / s_filename

        sample_index = 0
        with s_file_path.open('w') as file:
            while True:
                if debug:
                    sample_index += 1
                    print('Log: '
                          '[%s.%s] '
                          'sample: %d' % (UtilityBounding.__name__,
                                          UtilityBounding.rectangle_2_circle.__name__,
                                          sample_index)
                          )
                try:
                    """
                    get info
                    """
                    info = info_generator.__next__().strip().split(' ')

                    image_path = info.pop(0)

                    box_list = numpy.reshape(
                        numpy.array(info, dtype=numpy.float32),
                        [-1, 4]
                    )

                    """
                    convert to circle
                    """
                    circle_list = list()
                    for box in box_list:
                        centre_x = (box[0] + box[2]) / 2
                        centre_y = (box[1] + box[3]) / 2
                        radius = numpy.sqrt(
                            numpy.square(box[0] - box[2])
                            + numpy.square(box[1] - box[3])
                        ) / 2

                        circle = [centre_x, centre_y, radius]
                        circle_list.append(circle)

                    """
                    save to file
                    """
                    circles = numpy.reshape(
                        numpy.array(circle_list),
                        [-1]
                    )

                    new_info = image_path + ' ' + ' '.join(map(str, circles)) + '\n'
                    file.write(new_info)

                except StopIteration:
                    break

    @staticmethod
    def rectangle_2_square(rectangles,
                           image_width,
                           image_height):
        """
        convert boxes to square shape with avoiding exceeding original image by move bounding towards image
        :param rectangles:
        :param image_width:
        :param image_height:
        :return:
        """

        if rectangles is None:
            return None

        squares = numpy.copy(rectangles)

        boxes_height = rectangles[:, 3] - rectangles[:, 1] + 1
        boxes_width = rectangles[:, 2] - rectangles[:, 0] + 1

        long_side = numpy.maximum(boxes_height, boxes_width)

        squares[:, 0] = rectangles[:, 0] + 0.5 * boxes_width - 0.5 * long_side
        squares[:, 1] = rectangles[:, 1] + 0.5 * boxes_height - 0.5 * long_side

        squares[:, 2] = squares[:, 0] + long_side - 1
        squares[:, 3] = squares[:, 1] + long_side - 1

        invalid_indexes = numpy.less(squares[:, 0], 0)
        squares[:, 2] = numpy.where(invalid_indexes,
                                    squares[:, 2] - squares[:, 0],
                                    squares[:, 2])
        squares[:, 0] = numpy.where(invalid_indexes,
                                    0,
                                    squares[:, 0])

        invalid_indexes = numpy.less(squares[:, 1], 0)
        squares[:, 3] = numpy.where(invalid_indexes,
                                    squares[:, 3] - squares[:, 1],
                                    squares[:, 3])
        squares[:, 1] = numpy.where(invalid_indexes,
                                    0,
                                    squares[:, 1])

        invalid_indexes = numpy.greater_equal(squares[:, 2], image_width)
        squares[:, 0] = numpy.where(invalid_indexes,
                                    squares[:, 0] - (squares[:, 2] - image_width + 1),
                                    squares[:, 0])
        squares[:, 2] = numpy.where(invalid_indexes,
                                    image_width - 1,
                                    squares[:, 2])

        invalid_indexes = numpy.greater_equal(squares[:, 3], image_height)
        squares[:, 1] = numpy.where(invalid_indexes,
                                    squares[:, 1] - (squares[:, 3] - image_height + 1),
                                    squares[:, 1])
        squares[:, 3] = numpy.where(invalid_indexes,
                                    image_height - 1,
                                    squares[:, 3])

        return squares

    @staticmethod
    def iou_circle(circle,
                   circles):
        """
        calculate Intersection on Union of box against to each of boxes
        :param circle:
        :param circles:
        :return: intersection on union
        """

        # prepare
        circles = numpy.reshape(circles, [-1, 3])
        pi = 3.1415926
        iou_list = list()

        # get distance between circles
        distances = numpy.sqrt(
            numpy.square(circles[:, 0] - circle[0])
            + numpy.square(circles[:, 1] - circle[1])
        )

        for index, circle_2 in enumerate(circles):

            if circle_2[2] + circle[2] <= distances[index]:
                """
                separate
                """
                iou_list.append(0)

            elif numpy.abs(circle_2[2] - circle[2]) >= distances[index]:
                """
                include
                """
                if circle_2[2] > circle[2]:
                    g_radius = circle_2[2]
                    l_radius = circle[2]

                else:
                    g_radius = circle[2]
                    l_radius = circle_2[2]

                iou_list.append(
                    numpy.square(l_radius) / numpy.square(g_radius)
                )

            else:
                """
                intersect
                """
                # get rhombus_area
                perimeter = distances[index] + circle[2] + circle_2[2]

                rhombus_area = 2 * numpy.sqrt(
                    (perimeter / 2)
                    * (perimeter / 2 - circle_2[2])
                    * (perimeter / 2 - circle[2])
                    * (perimeter / 2 - distances[index])
                )

                # get sector area
                height = rhombus_area / distances[index]

                sector_area_1 = numpy.arcsin(height / circle_2[2]) * numpy.square(circle_2[2])
                sector_area_2 = numpy.arcsin(height / circle[2]) * numpy.square(circle[2])

                # get intersection
                inter_area = sector_area_1 + sector_area_2 - rhombus_area

                # get union
                union_area = pi * numpy.square(circle_2[2]) + pi * numpy.square(circle[2]) - inter_area

                # get iou
                iou = inter_area / union_area

                iou_list.append(iou)

        return iou_list

    @staticmethod
    def iou_rectangle(rectangle,
                      rectangles):
        """
        calculate Intersection on Union of box against to each of boxes
        :param rectangle:
        :param rectangles:
        :return: ious, inter_area, box_area, boxes_area
        """

        rectangles = numpy.round(numpy.reshape(rectangles, [-1, 4]))

        box_area = (rectangle[2] - rectangle[0] + 1) * (rectangle[3] - rectangle[1] + 1)
        boxes_area = (rectangles[:, 2] - rectangles[:, 0] + 1) * (rectangles[:, 3] - rectangles[:, 1] + 1)

        inter_x_1 = numpy.maximum(rectangle[0], rectangles[:, 0])
        inter_y_1 = numpy.maximum(rectangle[1], rectangles[:, 1])
        inter_x_2 = numpy.minimum(rectangle[2], rectangles[:, 2])
        inter_y_2 = numpy.minimum(rectangle[3], rectangles[:, 3])

        inter_w = numpy.maximum(0, inter_x_2 - inter_x_1 + 1)
        inter_h = numpy.maximum(0, inter_y_2 - inter_y_1 + 1)

        inter_area = inter_w * inter_h
        union = box_area + boxes_area - inter_area  # + 10e-10

        ious = inter_area / union

        return ious, inter_area, box_area, boxes_area

    UNION = 'Union'
    MIN = 'Minimum'

    @staticmethod
    def nms(boxes,
            scores,
            threshold,
            mode='Union',
            is_train=True):
        """
        non maximum suppression
        :param boxes:
        :param scores:
        :param threshold:
        :param mode:
        :param is_train:
        :return: kept indexes
        """

        x_1 = boxes[:, 0]
        y_1 = boxes[:, 1]
        x_2 = boxes[:, 2]
        y_2 = boxes[:, 3]

        boxes_area = (x_2 - x_1 + 1) * (y_2 - y_1 + 1)
        increase_order_indexes = numpy.argsort(-scores, axis=0)

        kept_indexes = list()

        while numpy.size(increase_order_indexes) > 0:
            index = increase_order_indexes[0]
            other_indexes = increase_order_indexes[1:]

            kept_indexes.append(index)

            if numpy.size(other_indexes) <= 0:
                break

            if is_train is False:
                # filter [1]:
                delta_x_1 = x_1[index] - x_1[other_indexes]
                delta_y_1 = y_1[index] - y_1[other_indexes]
                delta_x_2 = x_2[index] - x_2[other_indexes]
                delta_y_2 = y_2[index] - y_2[other_indexes]

                # box in other boxes
                mask_1 = numpy.where(delta_x_1 >= 0, 1, 0)
                mask_1 *= numpy.where(delta_y_1 >= 0, 1, 0)
                mask_1 *= numpy.where(delta_x_2 <= 0, 1, 0)
                mask_1 *= numpy.where(delta_y_2 <= 0, 1, 0)

                # box contain other boxes
                mask_2 = numpy.where(delta_x_1 <= 0, 1, 0)
                mask_2 *= numpy.where(delta_y_1 <= 0, 1, 0)
                mask_2 *= numpy.where(delta_x_2 >= 0, 1, 0)
                mask_2 *= numpy.where(delta_y_2 >= 0, 1, 0)

                mask = mask_1 + mask_2

                mask_index = numpy.where(mask == 0)[0]
                if numpy.size(mask_index) <= 0:
                    break
                other_indexes = other_indexes[mask_index]

            # intersection over union
            inter_x_1 = numpy.maximum(x_1[index], x_1[other_indexes])
            inter_y_1 = numpy.maximum(y_1[index], y_1[other_indexes])
            inter_x_2 = numpy.minimum(x_2[index], x_2[other_indexes])
            inter_y_2 = numpy.minimum(y_2[index], y_2[other_indexes])

            inter_width = numpy.maximum(inter_x_2 - inter_x_1 + 1, 0.0)
            inter_height = numpy.maximum(inter_y_2 - inter_y_1 + 1, 0.0)

            inter_area = inter_width * inter_height

            if mode == UtilityBounding.UNION:
                ious = inter_area / (boxes_area[index] + boxes_area[other_indexes] - inter_area)
            elif mode == UtilityBounding.MIN:
                ious = inter_area / numpy.minimum(boxes_area[index],
                                                  boxes_area[other_indexes])
            else:
                print('Error: '
                      '[%s.%s] '
                      'mns mode can only be ''%s'' or ''%s''.' % (UtilityBounding.__name__,
                                                                  UtilityBounding.nms.__name__,
                                                                  UtilityBounding.MIN,
                                                                  UtilityBounding.UNION))
                return None

            mask_index = numpy.where(ious <= threshold)[0]
            if numpy.size(mask_index) <= 0:
                break
            increase_order_indexes = other_indexes[mask_index]

        return kept_indexes

    @staticmethod
    def regress_rectangles(rectangles, normalized_offsets):
        """
        calibrate boxes by adding offsets
        :param rectangles:
        :param normalized_offsets:
        :return:
        """
        boxes_height = rectangles[:, 2] - rectangles[:, 0] + 1
        boxes_width = rectangles[:, 3] - rectangles[:, 1] + 1

        normalized_offsets[:, 0] *= boxes_width
        normalized_offsets[:, 1] *= boxes_height
        normalized_offsets[:, 2] *= boxes_width
        normalized_offsets[:, 3] *= boxes_height

        calibrated_rectangles = rectangles + normalized_offsets

        return calibrated_rectangles

# if __name__ == '__main__':
